from ._adams import AdamsObjectDetectionReader
from ._coco import COCOObjectDetectionReader
from ._opex import OPEXObjectDetectionReader
from ._roi import ROIObjectDetectionReader
from ._voc import VOCObjectDetectionReader
from ._yolo import YoloObjectDetectionReader
